@@ -22,7 +22,7 @@ args = ''
 
 base_option_list = (
     make_option('-v', '--verbosity', action='store', dest='verbosity', default='1',
-        type='choice', choices=['0', '1', '2', '3'],
+        type='choice', choices=['0', '1', '2', '3', 'v'],
         help="Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output"),
     make_option('--pythonpath',
         help="A directory to add to the Python path, e.g. '/home/myproject'."),
